/*
 * CADisplay.java
 *
 * Created on May 15, 2003
 */

/**
 * @version 1.0 
 * @author  Harry Fu
 */

import java.awt.*;
import javax.swing.*;

/**
 * CADisplay displays the world of Cellular Automata
 */
public class CADisplay extends JPanel {
   // number of gradient colors for CV CA
   private static final short CLR_SPACE = 360; 
   private int width = 0;              // width of the world
   private int height = 0;             // height of the world
   private int generationCt = 0;       // count for generation
   private int ruleCt = 0;             // count for rule K^N
   private int states = 0;             // count for state K
   private int neighbors = 0;          // count for neighbor N
   private int neighborRadius = 0;     // number of neighbor radius from center
   private int[] rule = null;          // rule count in array bit
   private int[] world = null;         // array of current world
   private int[] newWorld = null;      // array of new world
   private double[] cvOldWorld = null; // array of continous value old world
   private double[] cvWorld = null;    // array of continuous value world
   private double[] cvNewWorld = null; // array of continuous value new world
   private double cValue = 0.0D;       // current continuous value
   private double oldC = 0.0D;         // oldC
   private double LPlusR = 0.0D;       // L + R value 
   private Color[] stateColor = null;  // color for discrete CA
   private Color[] cvColor = null;     // color for continuous CA
   private double increment = 0.0D;    // increment value continous CA
   private double leftW = 0.0D;        // left weight for L cell
   private double rightW = 0.0D;       // right weight for R cell
   private Image offScrImage = null;   // offscreen image for CA display
   private CARule CARule = null;       // randomly generate a CA
   private CAColor caColor = null;     // color selection
   private boolean waveEquation = false; // apply wave equation
  
   /** Creates a new instance of CADisplay */
   public CADisplay() {
      states = 2;
      neighbors = 3;
      ruleCt = 8; 
      neighborRadius = 1;
      increment = 0.001D;
      leftW = 1;
      rightW = 1;
      caColor = new CAColor();
      CARule = new CARule();
   }
   /**
    * CAConfig use the current state count, neighbor count, rule array count, 
    * and state color.
    * @param stateCt count for state
    * @param neighborCt count for neighbor
    * @param curRule count for rule array
    */
   public synchronized void CAConfig(int stateCt, int neighborCt, int[] curRule, Color[] curColor) {
      states = stateCt;
      neighbors = neighborCt;
      neighborRadius = neighbors / 2;
      
      if (curRule != null)
         rule = curRule;
      else 
         rule = null;
      ruleCt = 1;
      for (int i = 0; i < neighbors; i++) 
         ruleCt *= states;
      if (stateColor != null && curColor.length >= states)
         stateColor = curColor;
      else
         curColor = null;
   }     
   /**
    * CVConfig use continous value CA rule.
    * @param a increment value for continous CA
    * @param b left weight for continous CA
    * @param c right weight for continous CA
    * @param curColor color choice for continous CA
    */
   public synchronized void CVConfig(double inc, double lw, double rw, Color[] curColor) {
      increment = inc;
      leftW = lw;
      rightW = rw;
      if (curColor != null)
         cvColor = curColor;
      else
         curColor = null;
   }   
   /**
    * setCAWorld set current CA world display.
    * @param randInit determine initial conidtion (one cell or random cells)
    */
   public synchronized void setCAWorld(String seedType) {  
      if ((width != getSize().width || height != getSize().height))
         offScrImage = null;
                  
      width = getSize().width;         // get current width
      height = getSize().height;       // get current height
      
      if ((world == null && newWorld == null) || 
            (world.length < width && newWorld.length < width)) {
         world = new int[width];          // set current world to current width
         newWorld = new int[width];       // set new world to current width
         cvOldWorld = new double[width];  // set old world buffer
         cvWorld = new double[width];     // set current CV world buffer
         cvNewWorld = new double[width];  // set new world buffer
      } 
      if (stateColor == null || stateColor.length < states) {
         stateColor = new Color[states]; // set state colors from number of state
         stateColor = caColor.getRainbowColor(states);
      }
      if (rule == null || rule.length < ruleCt) {
         rule = new int[ruleCt];
         rule = CARule.getRule();
      }
      for (int i = 0; i < width; i++) {
            world[i] = 0;              // initialize CA world array
            cvOldWorld[i] = 0.0D;      // initialize CV old world array
            cvWorld[i] = 0.0D;         // initialize CV world array
      }
      if (seedType.equals("One Seed")) { // start with one cell in center
         world [width/2] = 1;       
         cvWorld [width/2] = 1.0D;
         if (waveEquation)
            waveEquation = false;
      }
      else if (seedType.equals("Rand Seed")) {  // random start 
         System.arraycopy(CARule.getRandCASeed(width, states), 0, world, 0, width);
         System.arraycopy(CARule.getRandCVSeed(width), 0, cvWorld, 0, width);     
      }
      else if (seedType.equals("Linear Seed")) { // linear start 
         System.arraycopy(CARule.getLinearSeed(width, 35), 0, cvWorld, 0, width);
      }
      else if (seedType.equals("Sine Seed")) { // sine wave start 
         System.arraycopy(CARule.getSineSeed(width, 4), 0, cvWorld, 0, width);
      }      
      offScrImage = createImage(width, height);
      clearCAWorld();
   } // end setCAWorld()
   /**
    * clearCAWorld clears the graphic display.
    */
   public synchronized void clearCAWorld() {
      if (offScrImage != null) {
         Graphics g = offScrImage.getGraphics();
            if (stateColor != null)
               g.setColor(stateColor[0]);
            else
               g.setColor(stateColor[1]);
            g.fillRect(0, 0, width, height);
            g.dispose();
            repaint();
      } 
      else {
         Graphics g = getGraphics();
         if (stateColor != null) 
            g.setColor(stateColor[0]);
         else
            g.setColor(stateColor[1]);

         g.fillRect(0, 0, width, height);
         g.dispose();
      }
      generationCt = 0;        
   } // end clearCAWorld()
   /**
    * nextCAGeneration runs the next generation of discrete CA.
    */
   public synchronized void nextCAGeneration() {
      for (int i = 0; i < width; i++) {
         int L = i - neighborRadius;      // set L neighbor cell to left of center cell
         if (L < 0)                       // set wrap width for world
            L += width;                
         int R = i + neighborRadius;      // set R neighbor cell to right of center cell
         if (R >= width)                  // set wrap width for world
            R -= width;
         int rBit = 0;                    // rule bit 0 - 7 [111] [110] [101] [100] [011] [010] [001] [000]
         if (L < R) {                     // if L < R, determine the rule bit 0 - 7
         for (int j = L; j <= R; j++) 
            rBit = states * rBit + world[j];
         }
         else {
            for (int j = L; j < width; j++)
               rBit = states * rBit + world[j];
            for (int j = 0; j <= R; j++)
               rBit = states * rBit + world[j];
         }
         newWorld[i] = rule[rBit];           // set newWorld with new rule bit
      }
      displayCAWorld();
      System.arraycopy(newWorld, 0, world, 0, width); // optimize for speed than regular for loop
   } // end nextCAGeneration()
   /**
    * nextCVGeneration runs the next generation of continuous CA.
    */
   public synchronized void nextCVGeneration() {
      for (int i = 0; i < width; i++) {
         int L = i - 1;    // set L neighbor cell to left of center cell
         if (L < 0)        // set wrap width for world
            L += width;                
         int R = i + 1;    // set R neighbor cell to right of center cell
         if (R >= width)   // set wrap width for world
            R -= width; 
         cValue = 0;
         oldC = 0;
         LPlusR = 0;
         if (L < R) {                     
            for (int j = L; j <= R; j++) {
               /* lw*L + C + rw*R = cValue*/
               /* L + R = LPlusR; oldC*/
               if (j == L) {
                  cValue += (leftW * cvWorld[j]);     // L
                  LPlusR += (cvWorld[j] - increment); // L
               }
               if (j == R) {
                  cValue += (rightW * cvWorld[j]);    // R
                  LPlusR += (cvWorld[j] - increment); // R
               }
               if (j == (L+1)) {
                  oldC = (cvOldWorld[j]);             // OldC
                  cValue += (cvWorld[j]);             // C
               }
            }    
         }
         else {
            for (int j = L; j < width; j++) {
               /* lw*L + rw*R = cValue*/
               /* L + R = LPlusR; oldC*/
               if (j == L) {
                  cValue += (leftW * cvWorld[j]);     // L
                  LPlusR += (cvWorld[j] - increment); // L
               }
               if (j == (L+1)) {
                  oldC = (cvOldWorld[j]);             // OldC
                  cValue += (cvWorld[j]);             // C
               }
            }     
            for (int j = 0; j <= R; j++) {       
               /* lw*L + C + rw*R */
               /* L + R = LPlusR; oldC*/
               if (j == R) {       
                  cValue += (rightW * cvWorld[j]);    // R
                  LPlusR += (cvWorld[j] - increment); // R
               }
               if (j == (R-1)) {
                  oldC = (cvOldWorld[j]);             // OldC
                  cValue += (cvWorld[j]);             // C
               }
            }
         }
         if (waveEquation) // wave equation
            cValue = Math.abs(LPlusR - oldC) % 1.0;   // newC = (L + R) - OldC
         else
            cValue = ((cValue / 3) + increment) % 1;  // newC = [(C + L + R)/3  + a]  MOD 1
         cvNewWorld[i] = cValue;                      // translate rule bit to new world          
      }
      displayCVWorld();
      System.arraycopy(cvWorld, 0, cvOldWorld, 0, width); // save old buffer for wave equation
      System.arraycopy(cvNewWorld, 0, cvWorld, 0, width); // optimize for speed than regular for loop
   } // end nextCVGeneration()
   /**
    * display the world
    */   
   private synchronized void displayCAWorld() {
      Graphics g = null;
      if (offScrImage == null)
         g = getGraphics();
      else
         g = offScrImage.getGraphics();
      
      if (generationCt >= height && g != null) {
         g.copyArea(0, 0, width, height, 0, -1);
         generationCt = height - 1;
      }
      if (g != null) {
         int n = 0;
         for (int i = 0; i < width; i++) {   // draw one pixel at a time from left to right
            g.setColor(stateColor[world[i]]);
            g.fillRect(i, generationCt, 1, 1);  
         }
         generationCt++;
         g.dispose();
      }
      if (offScrImage != null) {             // draw image from offScrImage
         g = getGraphics();
         if (g != null && offScrImage != null) {
            g.drawImage(offScrImage, 0, 0, this);
            g.dispose();
         }
      } 
   } // end displayCAWorld()
   /**
    * display the CV world
    */   
   private synchronized void displayCVWorld() {
      Graphics g = null;
      if (offScrImage == null)
         g = getGraphics();
      else
         g = offScrImage.getGraphics();
      
      if (generationCt >= height && g != null) {
         g.copyArea(0, 0, width, height, 0, -1);
         generationCt = height - 1;
      }
      if (g != null) {
         for (int i = 0; i < width; i++) {   // draw one pixel at a time from left to right
            int n = (int)(cvWorld[i] * (CLR_SPACE - 1));  // mapping the gradient colors for continuous CA 
            g.setColor(cvColor[n]);
            g.fillRect(i, generationCt, 1, 1);
         }
         generationCt++;
         g.dispose();
      }
      if (offScrImage != null) {             // draw image from offScrImage
         g = getGraphics();
         if (g != null && offScrImage != null) {
            g.drawImage(offScrImage, 0, 0, this);
            g.dispose();
         }
      } 
   } // end displayCVWorld()
   /**
    * Avoid flickering effect.
    * @param g Graphics object.
    */
   public void update(Graphics g) {
      paint(g);
   }
   /**
    * paint procedure for update graphics.
    * @param g Graphics object.
    */
   public void paint(Graphics g) {
      if (offScrImage != null)
         g.drawImage(offScrImage, 0, 0, this);
   }
   /**
    * destryOffScrImg dispose the offScrImage for graphics.
    */
   public void disposeOffScrImg() {
      offScrImage = null;
   }
   /**
    * getDefaultRule returns the first randomly generated rule for this CA
    */
   public int[] getDefaultRule() {
      return rule;
   }
   /**
    * getDefaultRuleCount returns the current rule count for array 
    */
   public int getDefaultRuleCt() {
      return ruleCt;
   }
   /**
    * getInc returns the current increment value for continouse CA
    */
   public double getInc() {
      return increment;
   }
   /**
    * getLW returns the current left weight value for L cell
    */
   public double getLW() {
      return leftW;
   }
   /**
    * getRW returns the current right weight value for R cell
    */   
   public double getRW() {
      return rightW;
   }
   /**
    * setWaveEquation() apply wave equation
    * @param we true for wave equation
    */
   public void setWaveEquation(boolean we) {
      waveEquation = we;
   }
} // end CADisplay

