/*
 * CARule.java
 *
 * Created on May 15, 2003
 */

/**
 * @version 1.0 
 * @author  Harry Fu
 */

import java.util.Random;

/**
 * CARule Class sets the rule for CA 
 */
public class CARule {
    
   private int[] rule = null;       // store current CA rule
   private Random rand = null;      // store random generated number

   /** Creates a new instance of CARule */
   public CARule() {
      // initialize a CARule object        
      rule = null;
      rand = new Random();
   }
   /**
    * setRule sets user input rules for discrete valued CA
    * @param ruleCode decimal rule code
    */
   public void setRule(int states, int neighbors, int ruleCode) {         
      int ruleCount = 1;
      for (int i = 0; i < neighbors; i++)
         ruleCount = ruleCount * states;
      rule = new int[ruleCount];
      for (int i = 0; i < ruleCount; i++) 
         rule[i] = (ruleCode >> i) % states;
   }
   /*
    * setNewRule compute a new random rule base on current state and neighbors
    * @param states current state
    * @param neighbors current neighbor count
    */
   public void setRandRule(int states, int neighbors) {
      int ruleCount = 1;
      for (int i = 0; i < neighbors; i++)
         ruleCount = ruleCount * states;
      rule = new int[ruleCount]; 
      for (int i = 0; i < ruleCount; i++) {
         int r = (int)(rand.nextDouble() * states); 
         if (r >= states)
            r = 0;
         rule[i] = r;
      }
   }
   /*
    * setMutateRule mute rule bit
    * @param states current state
    * @param neighbord current neighbor count
    * @param mRule rule to be mutated
    */
   public void setMutateRule(int states, int neighbors, int[] mRule) {
      int ruleCount = 1;
      for (int i = 0; i < neighbors; i++)
         ruleCount = ruleCount * states;
      rule = new int[ruleCount]; 
      rule = mRule;
      for (int i = 0; i < ruleCount; i++) {
         // randomize the rule for a particular bit
         int r = (int)(rand.nextDouble() * states); 
         if (r >= states)
            r = 0;
         // randomize a particular bit with new rule
         int b = (int)(rand.nextDouble() * ruleCount); 
         if (b >= ruleCount)
            b = 0;
         rule[b] = r;
      }
   } 
   /**
    * setRandInc() compute a random increment
    */
   public double getRandInc() {
      double inc = rand.nextDouble();  // for increment value in continous CA
      if (inc < 0.001D)
         inc = 0.001D;
      return inc;
   }
   /**
    * getRandLW() compute a random left weight
    */
   public double getRandLW() {
      double lw = rand.nextDouble();   // for left weight value for L Cell 
         lw = lw + 0.85D;              // add more weight                    
      return lw;
   }
   /**
    * getRandRW() compute a random right weight
    */
   public double getRandRW() {
      double rw = rand.nextDouble();   // for increment value for R Cell
         rw = rw + 0.85D;              // add more weight 
      return rw;
   }
   /**
    * getPointSeed compute the value for the first row of world with one cell start
    * @param width the world width for the array
    * @param numNodes number of nodes
    */
   public double[] getOneSeed(int width) {
      double[] world = new double[width];
      for(int i = 0; i < width; i++)
          world[i] = 0.0D;
      world[(int)(width / 2.0D + 0.5D)] = 1.0D;
      return world;
   }   
   /**
    * getRandCASeed randomized first row for CA
    * @param width array size
    * @param states current states
    */
   public int[] getRandCASeed(int width, int states) {
      int[] world = new int[width];
      for (int i = 0; i < width; i++) { // for random initial condition 
         int ca = (int)(rand.nextDouble() * states);  // for CA
            if (ca >= states)
               ca = 0;
         world[i] = ca;
      }
      return world;
   }
   /**
    * getRandCVSeed randomized first row for CV
    * @param width array size
    */
   public double[] getRandCVSeed(int width) {
      double[] world = new double[width];
      for (int i = 0; i < width; i++) { // for random initial condition 
      double cv = rand.nextDouble();   // for CV
         if (cv < 0.001D)
            cv = 0.0D;
         world[i] = cv;
      }
      return world;
   }
   /**
    * getLinearSeed compute the value for the first row of world with linear start
    * @param width the world width for the array
    * @param numSteps number of nodes
    */
   public double[] getLinearSeed(int width, int numSteps) {
      int i;
      double dx = 1.0D / (double)numSteps;
      double val;
      int midpoint = (int)(width / 2.0D + 0.5D);
      double[] world = new double[width];
      for(i = 0; i < width; i++)
          world[i] = 0.0D;
      val = dx;    
      for(i = midpoint - numSteps + 1; i < midpoint; i++) {
          world[i] = val;
          val += dx;
      }
      val = 1.0D;
      for(; i < midpoint + numSteps; i++) {
          world[i] = val;
          val -= dx;
      }
      return world;
    }
   /**
    * getSineSeed compute the value for the first row of world with sine start
    * @param width the world width for the array
    * @param numNodes number of nodes
    */
   public double[] getSineSeed(int width, int numNodes) {
      double counter = 0.0D;
      double dx = 2.0D * Math.PI / (width / numNodes);
      double[] world = new double[width];
      for(int i = 0; i < width; i++)	
      {
         world[i] = (double)(Math.sin(counter) * 0.5D + 0.5D);
         counter = (counter + dx) % 360.0D;
      }
      return world;
   }
   /**
    * getRule gets current rule table
    */
   public int[] getRule() {
      return rule;
   }
} // end CARule
